{ Contains the code for all file input and output. All I/O should be done using
  these procedures. There is one exception. In the note.pas file the drivers are
  written to the driver file directly from the memo component. It is done that
  way because the memo component has a function to write to a file and using
  that function is easier then converting the text in the memo to the correct
  form for the writedriverfile function. }
unit fileio;
 
interface

uses sysutils, Dialogs, stypes, frontend;

type  TAction = (flRead, flWrite);

function  ReadParamFile(var filename:string; const npar:integer;
           var parray:paramarray; const nstat:integer;
           var sarray:statearray):Boolean;
procedure WriteParamFile(filename:string; const npar:integer;
          var parray:paramarray; const nstat:integer; var sarray:statearray);
procedure OpenDriverFile(filename:string; const numdrive:integer;
          var darray:drivearray; Action:TAction);
procedure GetCurrentDrivers(CurrTime:double; var CurrDrivers:DriveArray);
function  ReadDriverFile(var drivetime:double; const numdrive:integer;
          var darray:drivearray):Boolean;
function  WriteDriverFile(var drivetime:double; const numdrive:integer;
          var darray:drivearray):Boolean;
procedure CloseDriverFile;
procedure OpenOutputFile(filename:string;const numstat:integer;
          var sarray:statearray; const numproc:integer; var parray:processarray;
          Action:TAction);
function  ReadOutputFile(var time:double; const numdrive:integer; var darray:
          drivearray; const numstat:integer; var sarray:statearray; const
          numproc:integer; var parray:processarray):Boolean;
function WriteOutputFile(time:double; const numstat:integer;
          var sarray:statearray; const numproc:integer;
          var parray:processarray):Boolean;
procedure CloseOutputFile;
procedure OpenBatchFile(filename:string);
function  ReadBatchFile(filename:string; var paramname, drivername, outputname:string;
                        var begintime, endtime,steptime:double):Boolean;
procedure CloseBatchFile;
procedure OpenLogFile(filename, batchfile:string);
function  WriteLogFile(filename:string; outputname:string;
                        StatusMessage:string):Boolean;
procedure CloseLogFile;

var
 driverfile:textfile;
 outfile:textfile;
 batchfile:textfile;
 logfile:textfile;
 ParamFileVersion:string;

implementation


// Parameter File I/O

{ The parameter file is a space delimited ASCII file containing the parameter
  values and the initial values of the state variables. Each line of the file
  contains one parameter or state variable. The first item on the line is the
  variable name (a string up to 20 characters), the second item is the numerical
  value of the item and the final item on the line is the units of the parameter
  or state variable (also a string up to 20 characters long). }

{ This function reads the entire parameter file if it exists or creates a new
   empty file. }
function ReadParamFile(var filename:string; const npar:integer;
           var parray:paramarray; const nstat:integer;
           var sarray:statearray):Boolean;
var
  i,charnum:integer;
  dum,dum2:string[20];
  tempstring1,tempstring2:string;
  paramfile:textfile;
begin
  ReadParamFile := False;
  assignfile(paramfile,filename);
  if not FileExists(filename) then   // create a new parameter file
    begin
      NeedToSavePar := true;
      rewrite(paramfile);
      ReadParamFile := true;
    end
  else    // read existing parameter file
    begin
      NeedToSavePar := False;
      reset(paramfile);
      readln(paramfile,ParamFileVersion);
      SetLength(tempstring1,22);
      strlcopy(PChar(tempstring1),PChar(ParamFileVersion),21);
    { Delete any spaces in the model name read from the parameter file and any
      spaces in a copy of the modelname field of modeldef. This is to prevent
      the invalid parameter file version error due to differences in spacing. }
      charnum := pos(' ',tempstring1);
      while charnum <> 0 do
        begin
          delete(tempstring1,charnum,1);
          charnum := pos(' ',tempstring1);
        end;
    // And now delete the spaces from the modelname.
      tempstring2 := ModelDef.modelname;
      charnum := pos(' ',tempstring2);
      while charnum <> 0 do
        begin
          delete(tempstring2,charnum,1);
          charnum := pos(' ',tempstring2);
        end;

    // Compare the 2 strings containing the modelname.
      if StrIComp(PChar(tempstring1),PChar(tempstring2)) <> 0 then
       begin
        reset(paramfile);
        ParamFileVersion := 'Unknown'; // Model names don't match
       end
      else  // Model names do match, now extract the version number and compare
       begin
        delete(ParamFileVersion,1,21);
        charnum := pos(' ',ParamFileVersion);
        if charnum <> 0 then
          delete(ParamFileVersion,charnum,Length(ParamFileVersion)-charnum+1);
       end;

      if ParamFileVersion <> ModelDef.versionnumber then
       begin
        MessageDlg('Parameter File Version does not match Model Version. '
                    + 'No parameters read.', mtError,[mbOK],0);
        filename := '';
       end
      else
{ Model name and version match the parameter file so read in the state variables
  and parameters. }
       begin
        for i:= 1 to nstat do  // Read state variables ignoring names and units
         readln(paramfile,dum,sarray[i].value,dum2);
        for i:= 1 to npar do   // Read parameters ignoring names and units
         readln(paramfile,dum,parray[i].value,dum2);
        ReadParamFile := true;
       end;
    end;
  closefile(paramfile);
end;  // End readparamfile

{ Writes the entire parameter file, overwriting any existing file. Frontend
  does the checking to be sure you want to overwrite an existing file. See
  DlgSaveParam. }
procedure WriteParamFile(filename:string; const npar:integer;
           var parray:paramarray; const nstat:integer;
           var sarray:statearray);
var
  i:integer;
  paramfile:textfile;
begin
  assignfile(paramfile,filename);
  rewrite(paramfile);
  // Write out the model name and version.
  writeln(paramfile,format('%-20.20s %-20.20s',[ModelDef.modelname,
                                                ModelDef.VersionNumber]));
  // Write state variable name, value and units
  for i:= 1 to nstat do
    writeln(paramfile,format('%-20.20s',[sarray[i].name]),sarray[i].value,' ',
                             format('%-20.20s',[sarray[i].units]));
  // Write parameter name, value and units
  for i:= 1 to npar do
    writeln(paramfile,format('%-20.20s',[parray[i].name]),parray[i].value,' ',
                             format('%-20.20s',[parray[i].units]));
  closefile(paramfile);
  NeedToSavePar := false;
end;  // end of writeparamfile


// Driver File I/O

{ The driver file is a space delimited ASCII file containing the values for the
  driver variables at specified times. Each line of the file contains a value for
  all the drivers for the specified time. The first item on the line is the time,
  followed by the drivers. The first line of the file contains the names of the
  variables, the second contains the units of the items, and the third line
  contains the actual values of the drivers. Subsequent lines contain the drivers
  for later times. Driver variables are held at the last value read until the
  time specified on the next line. Negative times cause the drivers to be
  ramped from the values specified on the line before the negative time to the
  values specified on the line with the negative time. The function
  GetCurrentDrivers calculates any ramping if necessary. Access to the
  drivers while running should be done using GetCurrentDrivers.

  The driver file must be opened and closed using OpenDriverFile and
  CloseDriverFile.  Be sure to call CloseDriverFile when done. }

// Opens the driver file for reading or writing depending on the value of Action
procedure OpenDriverFile(filename:string; const numdrive:integer;
          var darray:drivearray; Action:TAction);
var
 tempstring:string;
 i:integer;
begin
 assignfile(driverfile,filename);
 if Action = flRead then  // Open driver file for reading
  begin
   if not FileExists(filename) then
    begin
     MessageDlg('Cannot read driver file. File does not exist.',
               mtError,[mbOK],0);
    end
   else   // File exists so read in the header lines
    begin
     reset(driverfile);
     readln(driverfile,tempstring);    // Read driver names and throw away
     readln(driverfile,tempstring);    // Read driver units and throw away
    end;
  end
 else
{ Open driver file for writing and write out driver names and units. }
  begin
    rewrite(driverfile);  // Create a new driver file
    write(driverfile,format('%-20.20s',['Time']),' ');  // Write time header
    for i := 1 to numdrive do        // Write driver names
     write(driverfile,format('%-20.20s',[darray[i].name]),' ');
    writeln(driverfile);   // Write return
    write(driverfile,format('%-20.20s',[ModelDef.timeunit]),' '); // Write time unit
    for i := 1 to numdrive do     // Write driver units
     write(driverfile,format('%-20.20s',[darray[i].units]),' ');
    writeln(driverfile);   // Write return
  end;
end;   // end of opendriverfile

{ This procedure gets the values for the drivers at the time specified in
  CurrTime. If the next_time read in from the driver file is negative, this
  procedure calculates a linear ramp between the values in last_drive and
  next_drive. }
procedure GetCurrentDrivers(CurrTime:double; var CurrDrivers:DriveArray);
var
 tfrac : double;
 i:integer;
begin
// No previous values have been read. Initialize everything.
 if (last_time = 0) and (next_time = 0) then
  begin
   ReadDriverFile(next_time,ModelDef.numdrive,next_drive);
   last_time := next_time;
   last_drive := next_drive;
   CurrDrivers := next_drive;
  end;

{ If CurrTime (passed to this procedure from the calling procedure) is greater
  than the value in next_time, read in new values for the drivers. }
 while CurrTime >= abs(next_time) do
  begin
   last_time := next_time;
   CurrDrivers := next_drive;
   last_drive := next_drive;
   if not ReadDriverFile(next_time,ModelDef.numdrive,next_drive)
       then next_time := 2*time_stop;
  end;

{ If next_time is less than zero calculate the drivers under a linear ramp
  between last_time and next_time. }
 if (next_time < 0) and  (abs(next_time)<>abs(last_time)) then
  begin
   tfrac := (CurrTime-abs(last_time))/(abs(next_time)-abs(last_time));
   for i := 1 to ModelDef.numdrive do
     CurrDrivers[i].value := last_drive[i].value + tfrac*(next_drive[i].value -
                                       last_drive[i].value);
  end;
end;    // getcurrentdrivers

// Reads one line of the driver file
function ReadDriverFile(var drivetime:double; const numdrive:integer;
                            var darray:drivearray):Boolean;
var
  j:integer;
  temptime:double;
  tempdrive:drivearray;
  EmptyLine:Boolean;
begin
// Copy the names and units from darray to tempdrive so they aren't overwritten.
   tempdrive := darray;
   if not eof(driverfile) then // Check for end of file
    begin
     read(driverfile,temptime);   // Read drive time
     for j:=1 to numdrive do
      read(driverfile,tempdrive[j].value);  // Read drivers
     readln(driverfile);      // Advance to next line
     EmptyLine := False;
// If all the drivers read in are zero then the line is an empty line.
     if (temptime = 0) then
       begin
        EmptyLine := True;
        for j := 1 to numdrive do
          EmptyLine := EmptyLine and (tempdrive[j].value = 0);
       end;
// Read successful assign the values read into darray and drivetime
     if not EmptyLine then
      begin
        drivetime := temptime;
        darray := tempdrive;
        ReadDriverFile := True;  // Read successful
      end
     else
// Read in an empty line, advance to the end of the file.
      begin
        while not eof(driverfile) do begin
          readln(driverfile);
        end;
        ReadDriverFile := False;
      end;
    end
   else
    ReadDriverFile := False;  // End of file, read unsuccessful
end;     // readdriverfile

// Writes one line to the driver file.
function WriteDriverFile(var drivetime:double; const numdrive:integer;
           var darray:drivearray):Boolean;
var
 i:integer;
begin
   write(driverfile,format('%-20.13e',[drivetime]),' ');   // Write drive time
   for i := 1 to numdrive do
    write(driverfile,format('%-20.13e',[darray[i].value]),' '); // Write drivers
   writeln(driverfile);   // Write return
   WriteDriverFile := True;
end;   // writedriverfile

// Closes the driver file
procedure CloseDriverFile ;
begin
 closefile(driverfile);
end;


// Output File I/O

{ The output file is a space delimited ASCII file containing values for the
  state variables and processes at each time step of the model run. The first
  item on a line is the time followed by the state variables and then the
  process variables. The first two lines of the file are header lines. The first
  line contains the names of the variables and the second line contains the
  units.

  The output file must be opened and closed using OpenOutputFile and
  CloseOutputFile. Be sure to call CloseOutputFile when you are done.}

// Opens the output file for reading or writing depending on the value of Action
procedure OpenOutputFile(filename:string;const numstat:integer;
          var sarray:statearray; const numproc:integer; var parray:processarray;
          Action:TAction);
var
  j:integer;
  temp:string;
begin
 assignfile(outfile,filename);
 if Action = flread then  // Open output file for reading
   begin
     reset(outfile);
     readln(outfile,temp); // Reads variable names and throws them away
     readln(outfile,temp); // Reads variable units and throws them away
   end
 else  // Open output file to write
   begin
    rewrite(outfile);    // Create a new output file
    write(outfile,format('%20.20s',['Time']),' ');  // Write Variable Names
   for j:= 1 to ModelDef.numdrive
      do write(outfile,format('%20.20s',[drive[j].name]),' ');
    for j:=1 to numstat do
        write(outfile,format('%20.20s',[sarray[j].name]),' ');
    for j:=1 to numproc do
        write(outfile,format('%20.20s',[parray[j].name]),' ');
    writeln(outfile);
    write(outfile,format('%20.20s',[ModelDef.timeunit]),' '); // Write Variable
   for j:= 1 to ModelDef.numdrive                              //   Units
      do write(outfile,format('%20.20s',[drive[j].units]),' ');
    for j:=1 to numstat do
        write(outfile,format('%20.20s',[sarray[j].units]),' ');
    for j:=1 to numproc do
        write(outfile,format('%20.20s',[parray[j].units]),' ');
    writeln(outfile);
   end;
end;  // openoutputfile

// Reads one line of the output file
function ReadOutputFile(var time:double; const numdrive:integer; var darray:
         drivearray; const numstat:integer; var sarray:statearray; const
         numproc:integer;var parray:processarray):Boolean;
var
 j:integer;
begin
   if eof(outfile) then
     ReadOutputFile := False
   else
    begin
     read(outfile, time);    // Read time
     for j:= 1 to ModelDef.numdrive
      do read(outfile,darray[j].value);
     for j := 1 to numstat
        do read(outfile,sarray[j].value); // State variables
     for j := 1 to numproc
        do read(outfile,parray[j].value); // Process variables
     readln(outfile);  // Advance to next line
     ReadOutputFile := True;
    end;
end;  // readoutputfile

// Writes one line to the output file
function WriteOutputFile(time:double; const numstat:integer;
         var sarray:statearray; const numproc:integer;
         var parray:processarray):Boolean;
var
  j:integer;
begin
   write(outfile,format('%20g',[time]),' ');  // Write time
   for j:= 1 to ModelDef.numdrive
      do write(outfile,format('%20.13e',[drive[j].value]),' ');
   for j:=1 to numstat     // State variables
      do write(outfile,format('%20.13e',[sarray[j].value]),' ');
   for j:=1 to numproc     // Process variables
      do write(outfile,format('%20.13e',[parray[j].value]),' ');
   writeln(outfile);  // Write return
   WriteOutputFile := True;
end;  // writeoutputfile

// Closes the output file
procedure CloseOutputFile;
begin
   closefile(outfile);
end;

procedure OpenBatchFile(filename:string);
var
tempstring:string;
begin
 assignfile(batchfile,filename);
 reset(batchfile);
 readln(batchfile,tempstring);    // Read names and throw away
end;

function  ReadBatchFile(filename:string; var paramname, drivername, outputname:string;
                        var begintime, endtime,steptime:double):Boolean;
var
 tempstring1,tempstring2: string;
 numbegin,numend:integer;
begin
 if not eof(batchfile) then // Check for end of file
  begin
   readln(batchfile,tempstring1);   // Read line
   numbegin := 1;
   numend := pos(',',tempstring1);
   paramname := copy(tempstring1,numbegin,numend-numbegin+1);
   delete(tempstring1,numbegin,numend);
   delete(paramname,pos(',',paramname),1);

   numend := pos(',',tempstring1);
   drivername := copy(tempstring1,numbegin,numend-numbegin+1);
   delete(tempstring1,numbegin,numend);
   delete(drivername,pos(',',drivername),1);

   numend := pos(',',tempstring1);
   outputname := copy(tempstring1,numbegin,numend-numbegin+1);
   delete(tempstring1,numbegin,numend);
   delete(outputname,pos(',',outputname),1);

   numend := pos(',',tempstring1);
   tempstring2 := copy(tempstring1,numbegin,numend-numbegin+1);
   delete(tempstring1,numbegin,numend);
   delete(tempstring2,pos(',',tempstring2),1);
   begintime := strtofloat(tempstring2);

   numend := pos(',',tempstring1);
   tempstring2 := copy(tempstring1,numbegin,numend-numbegin+1);
   delete(tempstring1,numbegin,numend);
   delete(tempstring2,pos(',',tempstring2),1);
   endtime := strtofloat(tempstring2);

   numend :=  length(tempstring1);
   tempstring2 := copy(tempstring1,numbegin,numend-numbegin+1);
   delete(tempstring1,numbegin,numend);
   delete(tempstring2,pos(',',tempstring2),1);
   steptime := strtofloat(tempstring2);
   ReadBatchFile := True;
  end
 else
   ReadBatchFile := False;
end;

procedure CloseBatchFile;
begin
 closefile(batchfile);
end;

procedure OpenLogFile(filename, batchfile:string);
begin
 assignfile(logfile,filename);
 rewrite(logfile);
 writeln(logfile, 'Output log for Modelshell Batch Utility Version ' + ModelDef.versionnumber);
 writeln(logfile, 'Generated: ' + DateTimeToStr(Now));
 writeln(logfile, 'Output from gem batch file: ' + batchfile);
end;

function  WriteLogFile(filename:string; outputname:string;
                        StatusMessage:string):Boolean;
begin
 writeln(logfile, outputname + ': ' + StatusMessage);
 WriteLogFile := True;
end;

procedure CloseLogFile;
begin
 closefile(logfile);
end;

end.
